<div align="center"><img height="180" src="https://gitlab.iscpif.fr/gargantext/main/raw/master/images/logo.png"></div>

&nbsp;
# Gargantext with Haskell (Backend instance)

![Haskell](https://img.shields.io/badge/Code-Haskell-informational?style=flat&logo=haskell&color=6144b3)&nbsp;&nbsp;![Nix](https://img.shields.io/badge/Package%20manager-Nix-informational?style=flat&logo=nixos&color=6586c8)&nbsp;&nbsp;![Cabal](https://img.shields.io/badge/Tools-Cabal-informational?style=flat&logo=cabal&color=567dd9)&nbsp;&nbsp;![Stack](https://img.shields.io/badge/Tools-Stack-informational?style=flat&logo=stack&color=6144b3)&nbsp;&nbsp;![GHC](https://img.shields.io/badge/Tools-GHC-informational?style=flat&logo=&color=2E677B)&nbsp;&nbsp;![Docker](https://img.shields.io/badge/Tools-Docker-informational?style=flat&logo=docker&color=003f8c)

#### Table of Contents
1. [About the project](#about)
2. [Installation](#install)
3. [Use cases](#use-cases)
4. [Notes for developers](#develop)
5. [`haskell-language-server`](#haskell-language-server)
6. [Running the tests](#running-tests)
7. [Async workers](#async-workers)
8. [Notifications](#notifications)

## About the project <a name="about"></a>

GarganText is a collaborative web-decentralized-based macro-service platform for the exploration of unstructured texts. It combines tools from natural language processing, text-data-mining bricks, complex networks analysis algorithms and interactive data visualization tools to pave the way toward new kinds of interactions with your textual and digital corpora.

This software is free (as "Libre" in French) software, developed by the CNRS Complex Systems Institute of Paris Île-de-France (ISC-PIF) and its partners.

GarganText Project: this repo builds the backend for the frontend server built by [backend](https://gitlab.iscpif.fr/gargantext/haskell-gargantext).


## Installation <a name="install"></a>

Disclaimer: since this project is still in development, this document remains in progress. Please report and improve this documentation if you encounter any issues.

### Prerequisites

You need to have the following installed:
* [GHCup](https://www.haskell.org/ghcup/)
* [Git](https://git-scm.com/book/en/v2/Getting-Started-Installing-Git)
* [Nix](https://nixos.org/download/)
* [Docker Compose](https://docs.docker.com/compose/install/).

You will need to add yourself as a member of the `docker` group for Docker
Compose to work (replace `myusername` with your username in the following
command):
```bash
sudo usermod -a -G docker myusername
```

### Getting the source code

```bash
git clone https://gitlab.iscpif.fr/gargantext/haskell-gargantext.git
cd haskell-gargantext
git clone https://gitlab.iscpif.fr/gargantext/haskell-gargantext.git
cd ..
```

### Building Gargantext

#### Building the Backend

Run the install script from inside the `haskell-gargantext/` directory:
```bash
./bin/install
```
This will build Gargantext, run the tests, and build the documentation.

#### Building the Frontend

Move into the `purescript-gargantext/` directory, then run the install script:
```bash
cd purescript-gargantext
./bin/install
cd ..
```

### Initialization

Rename the `gargantext-settings.toml_toModify` into `gargantext-settings.toml`,
and fill in the missing fields.

> `.gitignore` excludes this file, so you don't need to worry about committing
> it by mistake, and you can change the passwords in `gargantext-settings.toml`
> safely.

**NOTE** If you had the `gargantext.ini` file before, you can automatically
convert it into a file `gargantext-settings.toml` by running the following
from a Nix shell:
```shell
n$ ./bin/cli ini
```

### Launching and Initializing the Database

Launch the docker image from inside the dedicated directory:
```bash
cd devops/docker/
docker compose up
```

Then initialize the database using the dedicated command: from inside the
`haskell-gargantext/` directory, run
```bash
./bin/cli init
```

And provide a name and a password for the master user as requested.

### Running Gargantext

From inside the `haskell-gargantext/` directory, run
```bash
./bin/run
```

## Use Cases <a name="use-cases"></a>

### Multi-User with Graphical User Interface (Server Mode)

``` shell
$ ~/.local/bin/stack --docker exec gargantext-server -- --run Prod
```

Then you can log in with `user1` / `1resu`


### Command Line Mode tools

#### Simple cooccurrences computation and indexation from a list of Ngrams

``` shell
$ stack --docker exec gargantext-cli -- CorpusFromGarg.csv ListFromGarg.csv Ouput.json
```

### Analyzing the ngrams table repo

We store the repository in directory `repos` in the [CBOR](https://cbor.io/) file format. To decode it to JSON and analyze, say, using [jq](https://shapeshed.com/jq-json/), use the following command:

``` shell
$ cat repos/repo.cbor.v5 | stack exec gargantext-cbor2json | jq .
```
### Documentation

To build documentation, run:

```shell
$ stack build --haddock --no-haddock-deps --fast
```

(in `.stack-work/dist/x86_64-linux-nix/Cabal-3.2.1.0/doc/html/gargantext`).


## Notes for developers <a name="develop"></a>

### Developer Documentation

If you want to contribute to Gargantext, there is [some documentation geared to developers](https://write.frame.gargantext.org/s/805e1ee2bae24079554d24cdbc8ef5ba5c4ef7d83218c1e86c84af8ac269031d).

### Upgrading Haskell Packages

We use `gargantext.cabal`, `cabal.project` and `cabal.project.freeze`
as the source of truth. Ouf ot that, we generate the `stack.yaml` file
for those who prefer to use Stack.

Upgrading packages can be a pain sometimes, with cabal.

Here are some tips:

- Manually remove entries from your `cabal.project.freeze` to make the
  build a bit more "elastic";
- Lock the hackage-index state in the `cabal.project`, so that the
  solver won't try to pull newer dependencies;
- Specify constraints you want directly when building like `cabal
  v2-build --constraint tasty==x.y.z.w`
- Generate another `.freeze` with `cabal v2-freeze` once you got the
  new build to compile (this is good for small, incremental upgrades)
- Bounds in `.cabal` are definitely respected, but ofc the `.freeze`
  takes priority, so you want to maybe use `cabal gen-bounds` when
  your `.freeze` still exists, remove the file, try again.

Also, it's helpful to build with `stack build` from time to time. The
warnings are displayed, whenever a different stack lts package is used
than the one defined in `.cabal` file - it's an incentive to upgrade
the `.cabal` file versions.

Occasionally, you can get issues with the `allow-newer: *` constraint
from `cabal.project`. E.g. when I was building with GHC 9.4.7, I had
errors with `hashable-1.5.0`. The solution is:
```shell
cabal v2-build --constraint hashable==1.4.3.0
```
(we don't depend on `hashable` directly, but `allow-newer: *` is so
liberal that a package that is too new is used).

Overall, it's preferred to specify strict constraints in
`gargantext.cabal` file and to do that, one can use `stack ls dependencies`
to have an idea what works.

If you want to see the detailed build info for a given dependency:
```shell
cabal v2-build -v servant-server
```

Also, you might use the `-Wunused-packages` GHC option, to get a
warning about unused packages (make sure though you build all targets
with `cabal v2-build all`).

Also, here is a relevant discussion:
https://discourse.haskell.org/t/whats-your-workflow-to-update-cabal-dependencies/9475

### Running Tests

From nix shell:

```shell
n$ cabal v2-test --test-show-details=streaming
```

Or, from "outside":
```shell
$ nix-shell --run "cabal v2-test --test-show-details=streaming"
```

If you want to run particular tests, use (for Tasty):
```shell
cabal v2-test garg-test-tasty --test-show-details=streaming --test-option=--pattern='/job status update and tracking/
```
or (for Hspec):
```shell
cabal v2-test garg-test-hspec --test-show-details=streaming --test-option=--match='/Dispatcher, Central Exchange, WebSockets/'
```

### CI

The CI is on gitlab and is controlled by the `.gitlab-ci.yml` file.

The Docker image that is used can be built with:
```shell
podman build -t cgenie/gargantext:9.4.8 -f ./devops/docker/Dockerfile --format=docker .
```
or
```shell
docker build -t cgenie/gargantext:9.4.8 -f ./devops/docker/Dockerfile .
```

NOTE: if podman/docker complains about not enough space, it's probably
about tmpfs. In that case you can set
```shell
mkdir _build
export TMPDIR=$(pwd)/_build
```
(https://stackoverflow.com/questions/75845647/changing-podman-build-to-path-with-more-space).

### Working on libraries

When a development is needed on libraries (for instance, the HAL crawler in https://gitlab.iscpif.fr/gargantext/crawlers):

1. Ongoing devlopment (on local repo):
   1. In `cabal.project`:
      - add `../hal` to `packages:`
      - turn off (temporarily) the `hal` in `source-repository-package` 
   2. When changes work and tests are OK, commit in repo `hal`
2. When changes are commited / merged:
   1. Get the hash id, and edit `cabal.project` with the **new commit id**
   2. run `./bin/update-project-dependencies`
      - get an error that sha256 don't match, so update the `./bin/update-project-dependencies` with new sha256 hash
      - run again `./bin/update-project-dependencies` (to make sure it's a fixed point now)

> Note: without `stack.yaml` we would have to only fix `cabal.project` -> `source-repository-package` commit id. Sha256 is there to make sure CI reruns the tests.

### GraphQL

Some introspection information.

Playground is located at http://localhost:8008/gql

#### List all GraphQL types in the Playground

```
{
  __schema {
    types {
      name
    }
  }
}
```

#### List details about a type in GraphQL

```
{
  __type(name:"User") {
  	fields {
    	name
      description
      type {
        name
      }
  	}
	}
}
```
### PostgreSQL <a name="pgsql"></a>

#### Upgrading using Docker

https://www.cloudytuts.com/tutorials/docker/how-to-upgrade-postgresql-in-docker-and-kubernetes/

To upgrade PostgreSQL in Docker containers, for example from 11.x to 14.x, simply run:
```shell
$ docker exec -it <container-id> pg_dumpall -U gargantua > 11-db.dump
```

Then, shut down the container, replace `image` section in `devops/docker/docker-compose.yaml` with `postgres:14`. Also, it is a good practice to create a new volume, say `garg-pgdata14` and bind the new container to it. If you want to keep the same volume, remember about removing it like so:
```shell
$ docker-compose rm postgres
$ docker volume rm docker_garg-pgdata
```

Now, start the container and execute:
```shell
$ # need to drop the empty DB first, since schema will be created when restoring the dump
$ docker exec -i <new-container-id> dropdb -U gargantua gargandbV5
$ # recreate the db, but empty with no schema
$ docker exec -i <new-container-id> createdb -U gargantua gargandbV5
$ # now we can restore the dump
$ docker exec -i <new-container-id> psql -U gargantua -d gargandbV5 < 11-db.dump
```

#### Upgrading using 

There is a solution using pgupgrade_cluster but you need to manage the clusters version 14 and 13. Hence here is a simple solution to upgrade.

First save your data:
```shell
$ sudo su postgres
$ pg_dumpall > gargandb.dump
```

Upgrade postgresql:
```shell
$ sudo apt install postgresql-server-14 postgresql-client-14
$ sudo apt remove --purge postgresql-13
```
Restore your data:
```shell
$ sudo su postgres
$ psql < gargandb.dump
```

Maybe you need to restore the gargantua password
```shell
$ ALTER ROLE gargantua PASSWORD 'yourPasswordIn_gargantext-settings.toml'
```

Maybe you need to change the port to 5433 for database connection in your gargantext.ini file.

## `haskell-language-server` <a name="haskell-language-server"></a>

If you want to use `haskell-language-server` for GHC 9.4.7, install it
with `ghcup`:
```shell
ghcup compile hls --version 2.7.0.0 --ghc 9.4.7
```
https://haskell-language-server.readthedocs.io/en/latest/installation.html


## Running the tests <a name="running-tests"></a>

Running the tests can be done via the following command:

```hs
cabal v2-test --test-show-details=streaming --flags 'test-crypto no-phylo-debug-logs'
```

The flags have the following meaning:

* `test-crypto`: Switch to use very fast (but not production-secure) cryptography, so that tests runs
  faster;
* `no-phylo-debug-logs`: Suppresses the debugging logs which would normally be present in phylo pure (!) code.

In order for some tests to run (like the phylo ones) is **required** to install the `gargantext-cli` via:

```hs
cabal v2-install gargantext:exe:gargantext-cli
```

### Modifying a golden test to accept a new (expected) output

Some tests, like the Phylo one, use golden testing to ensure that the JSON Phylo we generate is
the same as an expected one. This allows us to catch regressions in the serialisation or in the algorithm.
Sometimes, however, we genuinely want to modify the output so that it's the new reference (i.e. the new
golden reference). To do so, it's enough to run the testsuite passing the `--accept` flag, for example:

```hs
cabal v2-test garg-test-tasty --test-show-details=streaming --flags 'test-crypto no-phylo-debug-logs' --test-option=--pattern='/Phylo/' --test-option=--accept"
```

# Async workers <a name="async-workers"></a>

Async workers allow us to accept long-running jobs and execute them
asynchronously. Workers can be spawned on multiple machines, which
allows for scaling.

To run the worker, follow these steps:
- start a PostgreSQL DB, usually the one with Gargantext DB is enough
- `"simple"` worker definition is in `gargantext-settings.toml`
- run worker: `cabal v2-run gargantext-cli -- worker --name simple`

When running the worker for the first time (or sending a job), best
attempt is made to ensure the DB exists (if not, we will try to create
it) and the `pgmq` schema is initialized. This allows for hassle-free
maintenance and easier onboarding.

The project that we base our worker is
[haskell-bee](https://gitlab.iscpif.fr/gargantext/haskell-bee/). It's
a more generic framework for managing asynchronous workers, supporting
different brokers. Here, we decided to use `pgmq` because we already
have PostgreSQL deployed.

## Design

Thanks to the fact that we already use `Servant.Jobs` (which executes
the jobs in the gargantext-API process), we can migrate our jobs more
easily to a different backend.

All job definitions are in `G.C.W.J.Types` under `Job`
datatype.

If you want to add a new job, just add a new constructor to `Job`
(with all the arguments this job needs), implement to/from JSON
serialization (so we can send and read that job via the broker) and
implement appropriate handler in `G.C.Worker` -> `performAction`.

## No storage backend

When you compare `haskell-bee` with the [celery
project](https://docs.celeryq.dev/en/stable/), you can notice 2
things:
- both have configurable brokers
- `haskell-bee` doesn't have a storage backend This is because, when
analyzing our tasks, we actually don't store any **task** results
anywhere. The progress can be reported immediately via notifications,
same for errors. And when a task is executed (e.g. add to corpus with
query), the garg db state is mutated.

One could discuss if such a **task** storage backend is needed: if you
want to debug some task, just run the worker and log its results. You
can inspect later the errors in the log. On the other hand, if we had
a production worker, which processes lots of tasks, it could be a pain
to store the results of all of them, as we mostly don't care about the
finished ones.

Also, the `haskell-bee` framework allows to add custom hooks to the
worker. In particular, search for `onJobError`/`onJobTimeout` in
`Worker.State`. We could trigger some `IO` action on these hooks
(logging, sending mail, firing rockets).


# Notifications <a name="notifications"></a>

The notifications mechanism has 3 basic components: the `central
exchange`, the `dispatcher` and the `websocket server`.

## Central exchange

This is the first entrypoint for notifications. The code lies in
`G.C.N.CentralExchange`. To create a central exchange, just call
`gServer` from that module. Notice that currently there can be 1
central exchange, but in principle there could be many.

The purpose of central exchange is to be **quick**: accept as many
notifications as possible (they can be sent via
`G.C.N.CentralExchange` -> `notify`), then forward them to the
`dispatcher`.

You can send only predefined messages, see `G.C.N.CE.Types` ->
`CEMessage`.

Please note that the central exchange acts as a proxy and, as such, it
can run as a separate linux process. Currently, though, it is just a
GHC thread, part of `gargantext-server`.

## Dispatcher

The code for the dispatcher lies in `G.C.N.Dispatcher`. To spawn the
server, use `withDispatcher`.

The dispatcher is responsible for subscribing clients to messages and
dispatching incoming messages to appropriate clients. Also, it
implements a throttling mechanism so if it is bombarded with lots of
equal messages, it will send only some of them to the client
(c.f. `sendDataMessageThrottled`).

The basic data structure associated with the dispatcher is a set of
`Subscription`'s. This is a `TVar`, i.e. a global mutable variable. An
incoming subscription is registered in this set, along with topics
(see `G.C.N.D.Types` -> `Topic` datatype).

For example, a client can say: "i want to listen to update tree events
under node X", which translates to subscribing to a `UpdateTree X`
topic.

Or, the client might want to listen to progress of an async task with
id `X`, then it's topic `UpdateWorkerProgress JobInfo`.

When a message is forwarded to the dispatcher (preferably from the
central exchange), the dispatcher then associates that message with
all subscriptions (see `G.C.N.Dispatcher` -> `ceMessageSubPred` and
the filtering associated with that predicate function) and sends an
appropriate message to them.

## WebSockets

In theory, clients could subscribe to the dispatcher by different
means, but currently the only implemented way is by connecting to the
websocket server (which lies under `/ws` API endpoint).

The implementation of the websocket server lies under
`G.C.N.Dispatcher.WebSocket`. It is closely related with the
`Subscription` datatype. Each incoming connection is registered as a
separate `Subscription`. A user can optionally authenticate himself,
his subscription will then be "updated" to allow to listen for node id
changes that he has permissions to (this is still a bit on TODO).

A disconnection results in a removed subscription.

There are also `subscribe` and `unsubscribe` requests, see the
`WSRequest` datatype in `G.C.N.Dispatcher.Types` for possible requests
that you can make to the websocket endpoint.

When the dispatcher sends a message, it reuses that websocket
connection (it's stored under the `Subscription` datatype).

**Important** Because dispatcher is so closely related to the
websockets (i.e. it reuses the same connection), it has to run in the
same process as websockets (in a different thread). Our API is
monolithic, i.e. all services are bundled together into one
`gargantext-server` process. However, if needed, we could split the
API from the websocket/dispatcher and have 2 different linux services
running.

## Overall architecture remarks

All connections to central exchange and dispatcher are made via
[nanomsg](https://nanomsg.org/). You can think of nanomsg as
"upgraded" TCP sockets. nanomsg helps us, in particular, in
reconnection and buffering. This is an implementation detail and if
we're not satisfied with how it works, we could change to something
different, e.g. the `LISTEN`/`NOTIFY` mechanism in postgresql (we use
postgresql anyways).

My main idea in chosing nanomsg was having something reliable, which
at the same time, didn't require maintenance (e.g. setting up another
service just for notifications).

## Configuration

See `gargantex-settings.toml_toModify`. Overall, the config for
notifications is quite simple:
```toml
[notifications]
central-exchange = { bind = "tcp://*:5560", connect = "tcp://localhost:5560" }
dispatcher = { bind = "tcp://*:5561", connect = "tcp://localhost:5561" }
```
The `bind`/`connect` strings are just forwarded to nanomsg, hence we
could switch `tcp` to `ipc` (but this limits us to a single machine)
or `inproc` (this limits us to inter-process communication).

The `bind` part is for the server, the `connect` part is for the
clients connecting to that server.

## Async workers

A particular use-case of notifications is the async worker mechanism
described above.

In particular, async workers don't use a storage backend to store task
progress and this has consequences for notifications.

You can't just ask the API: "give me all tasks for this node" anymore
(this was possible with servant-jobs). Instead, what the frontend
does, is it sends an "update tree" subscription for nodes in the tree.

Take a look at `G.C.W.J.Types` -> `getWorkerMNodeId`: it is a function
that, given a job, returns a possible `NodeId` associated with that
task (not all tasks need a `NodeId` though).

If the dispatcher receives an `UpdateWorkerProgress` message, it
checks the associated `NodeId` and finds `UpdateTree` subscriptions
associated with that node id. It then sends 2 messages: one for the
"normal" subscription to the `UpdateWorkerProgress` topic (the
frontend needs to know the job id already) and another
`UpdateWorkerProgress` to all those being subscribed to `UpdateTree`
for that node id.

This way, the frontend is able to be informed of new jobs that appear
under a node, and render the progress bar accordingly.

This "push" approach has many benefits:
- if server is idle, we don't keep polling for new jobs
- when user refreshes the page, even when his jobs are running, the
  frontend is able to quickly discover new jobs and there is no need
  to store the jobs id in local storage as we did previously (it's
  important though that jobs send a progress notification quite often,
  otherwise the job looks stale and isn't that quickly discoverable)
- we can propagate job progress not only to the user who created it,
  but to all users having access to that node (so to all users of a
  shared node, for example) without any effort
- if the API server is restarted (and the workers do keep running in
  the background), the frontend keeps reconnecting to WS. Upon
  connection, all node subscriptions are resent and the user can pick
  up worker progress quickly

Of course, everything has it's bad sides:
- we have to watch progress updates and decide, based on `remaining`
  count, that the job is finished (however, async workers has the
  `onJobFinish` hook - c.f. `G.C.Worker` - and we could send a special
  message about a finished job, if needed)
- as noted above, the progress messages are sent twice (via
  `UpdateTree` and via `UpdateWorkerProgres` subscriptions)
- you need to wait a while to discover running jobs and if they're
  stale or don't report very often, you won't know what's being
  processed currently
  
Still, I prefer the "push" approach, because it more closely mimicks
the "broker" design. Thing is, a broker is not a database: you should
just expect it to deliver your message and forget about it.

If need arises, we could create a postgresql table which stores task
progress/task results, but it's just another burden to keep up with
(consider failing tasks, repeated ones etc).
